import typing
import json
from dataclasses import dataclass
from abc import ABC, abstractmethod
from commonfate_provider import diagnostics
from commonfate_provider.dataclass import ModelMeta


@dataclass
class Field:
    usage: typing.Optional[str] = None
    secret: bool = False


class String(Field):
    def set(self, val: str) -> None:
        self._value = val

    def get(self) -> str:
        return self._value


class ConfigLoader(ABC):
    @abstractmethod
    def load(self) -> dict:
        pass


class StringLoader(ConfigLoader):
    def __init__(self, config_json: str) -> None:
        self.config = config_json

    def load(self):
        return json.loads(self.config)


class DictLoader(ConfigLoader):
    def __init__(self, config_dict: dict) -> None:
        self.config = config_dict

    def load(self):
        return self.config


class NoopLoader(ConfigLoader):
    def load(self):
        return {}


class MethodNotImplemented(Exception):
    pass


@dataclass
class GrantResult:
    access_instructions: typing.Optional[str] = None
    """
    Instructions on how to access the entitlements.
    Common Fate will display these to the user upon a successful Access Request.
    """


class Provider(ABC):
    def __init__(self, config_loader: ConfigLoader) -> None:
        self._internal_key = "default"
        config_dict = config_loader.load()
        all_vars = [
            (k, v) for (k, v) in vars(self.__class__).items() if not k.startswith("__")
        ]
        for k, v in all_vars:
            if isinstance(v, String):
                val = config_dict[k]
                v.set(val=val)
                setattr(self, k, v)

    def validate_config(self) -> dict:
        results = {}
        all_validators = _ALL_CONFIG_VALIDATORS.get(self._internal_key, {})
        for validator in all_validators.values():
            diags = diagnostics.Logs()
            try:
                validator.func(self, diags)
            except Exception as e:
                diags.error(str(e))

            results[validator.id] = {
                "logs": [l.__dict__ for l in diags.logs],
                "success": diags.succeeded(),
            }
        return results

    @classmethod
    def export_schema(cls) -> dict:
        """
        Exports config variables defined in a Provider class
        to a dictionary.

        For example, if a Provider is defined as:
        ```
        class Provider(provider.Provider):
            url = provider.String()
        ```

        this method will produce a dictionary in the form:
        ```
        {"url": {"type": "string"}}
        ```
        """
        config_vars = {}
        all_vars = [(k, v) for (k, v) in vars(cls).items() if not k.startswith("__")]
        for k, v in all_vars:
            if type(v) == String:
                config_vars[k] = {"type": "string", "usage": v.usage}
        return config_vars


def capabilities(_internal_key: str = "default") -> dict:
    return {"builtin": {}}


ConfigValidatorFunc = typing.Callable[[typing.Any, diagnostics.Logs], None]


@dataclass
class ConfigValidator:
    name: str
    id: str
    func: ConfigValidatorFunc


_ALL_CONFIG_VALIDATORS: typing.Dict[str, typing.Dict[str, ConfigValidator]] = {}
"""dict in the format {'default': {'id': {...}}"""


def config_validator(
    name: str,
    _internal_key: str = "default",
) -> typing.Callable[[ConfigValidatorFunc], ConfigValidatorFunc]:
    def actual_decorator(func: ConfigValidatorFunc):
        id = func.__name__
        cv = ConfigValidator(id=id, name=name, func=func)
        _ALL_CONFIG_VALIDATORS.setdefault(_internal_key, {})
        _ALL_CONFIG_VALIDATORS[_internal_key][id] = cv
        return func

    return actual_decorator


GrantValidatorFunc = typing.Callable[[typing.Any, str, typing.Any], None]


@dataclass
class GrantValidator:
    name: str
    id: str
    func: GrantValidatorFunc


_ALL_GRANT_VALIDATORS: typing.Dict[str, typing.Dict[str, GrantValidator]] = {}


def grant_validator(
    name: str,
    _internal_key: str = "default",
) -> typing.Callable[[GrantValidatorFunc], GrantValidatorFunc]:
    def actual_decorator(func: GrantValidatorFunc):
        id = func.__name__
        cv = GrantValidator(id=id, name=name, func=func)
        _ALL_GRANT_VALIDATORS.setdefault(_internal_key, {})
        _ALL_GRANT_VALIDATORS[_internal_key][id] = cv
        return func

    return actual_decorator


GrantFunc = typing.Callable[[typing.Any, str, typing.Any], typing.Optional[GrantResult]]
_GRANT: typing.Dict[str, GrantFunc] = {}


def _get_grant_func(_internal_key: str = "default"):
    return _GRANT[_internal_key]


def grant(
    _internal_key: str = "default",
) -> typing.Callable[[GrantFunc], GrantFunc]:
    # _internal_key is used for testing purposes, and allows
    # multiple independent decorators to be registered in a single test file
    def actual_decorator(func: GrantFunc):
        _GRANT[_internal_key] = func
        return func

    return actual_decorator


RevokeFunc = typing.Callable[[typing.Any, str, typing.Any], None]


_REVOKE: typing.Dict[str, RevokeFunc] = {}


def revoke(
    _internal_key: str = "default",
) -> typing.Callable[[RevokeFunc], RevokeFunc]:
    # _internal_key is used for testing purposes, and allows
    # multiple independent decorators to be registered in a single test file
    def actual_decorator(func: RevokeFunc):
        _REVOKE[_internal_key] = func
        return func

    return actual_decorator


def _get_revoke_func(_internal_key: str = "default"):
    return _REVOKE[_internal_key]



import abc
from typing_extensions import dataclass_transform
from dataclasses import dataclass


@dataclass_transform()
class ModelMeta(type):
    def __init_subclass__(
        cls,
        *,
        init: bool = True,
        frozen: bool = False,
        eq: bool = True,
        order: bool = True,
    ):
        return dataclass(cls)


class AbstractModelMeta(ModelMeta, abc.ABCMeta):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['commonfate_provider',
 'commonfate_provider.runtime',
 'commonfate_provider.runtime.tests']

package_data = \
{'': ['*']}

install_requires = \
['click==8.1.3',
 'pydantic==1.10.2',
 'pytest==7.2.0',
 'toml==0.10.2',
 'treelib==1.6.1',
 'typing-extensions==4.4.0']

setup_kwargs = {
    'name': 'commonfate-provider',
    'version': '0.1.11',
    'description': 'This is the core Provider framework on top of which Access Provider are written.',
    'long_description': '# commonfate-provider-core\nCommonfate Provider Core Python Package\n',
    'author': 'Chris Norman',
    'author_email': 'chris@commonfate.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import typing
from commonfate_provider.dataclass import ModelMeta
from commonfate_provider import tasks
from pydantic import BaseModel, Field
import inspect


def composite_id(fields: typing.List[str]):
    return "/".join(fields)


class Resource(BaseModel):
    id: str

    # def __init__(self, **data: typing.Any) -> None:
    #     setattr(self, "__dict__", data)

    def export_json(self) -> dict:
        return {"type": self.__class__.__name__, "data": dict(self)}

    # def __eq__(self, other: typing.Any) -> bool:
    #     if self.__class__ != other.__class__:
    #         return False
    #     return all(self.__dict__[k] == other.__dict__[k] for k in self.__dict__)

    # def __repr__(self) -> str:
    #     vals = [f"{k}={v}" for (k, v) in self.__dict__.items()]
    #     return self.__class__.__name__ + "(" + ", ".join(vals) + ")"


T = typing.TypeVar("T", bound=Resource)


class Query(typing.Generic[T]):
    def __init__(self, cls: typing.Type[T]) -> None:
        self.cls = cls

    def all(self) -> typing.List[T]:
        return DEFAULT_STORAGE.all(cls=self.cls)


def query(cls: typing.Type[T]) -> Query[T]:
    return Query(cls)


def Related(to: typing.Union[typing.Type[T], str]) -> str:
    if inspect.isclass(to):
        to = to.__name__
    return Field(relatedTo=to)


def Name() -> str:  # type: ignore
    pass


def UserEmail() -> str:  # type: ignore
    pass


class Context(metaclass=ModelMeta):
    def __init__(self, **data: typing.Any) -> None:
        setattr(self, "__dict__", data)


_ALL_RESOURCES: typing.List[Resource] = []


def _reset():
    global _ALL_RESOURCES
    _ALL_RESOURCES = []


def get() -> typing.List[Resource]:
    return _ALL_RESOURCES


def set_fixture(resources: typing.List[Resource]):
    global DEFAULT_STORAGE
    DEFAULT_STORAGE.resources = [r.export_json() for r in resources]


def fetcher(func: tasks.LoaderFunc):
    tasks._RESOURCE_LOADERS[func.__name__] = func
    return func


def audit_schema():
    """
    Returns the schema for the 'audit' section in a provider schema as a dict.
    This section defines the resources the provider can read, along with the
    fetching methods which can be called to fetch them.
    """
    loaders = {}
    for k in tasks._RESOURCE_LOADERS.keys():
        loaders[k] = {"title": k}

    resources = {}
    for Klass in Resource.__subclasses__():
        properties = {}
        all_vars = [(k, v) for (k, v) in vars(Klass).items() if not k.startswith("__")]
        for k, v in all_vars:
            properties[k] = {"type": "string"}
        resources[Klass.__name__] = Klass.schema()

    return {"resourceLoaders": loaders, "resources": resources}


def register(resource: Resource):
    _ALL_RESOURCES.append(resource)


def without_keys(d, keys):
    return {x: d[x] for x in d if x not in keys}


class JSONStorage:
    def __init__(self, resources: typing.List[dict]) -> None:
        self.resources = resources

    def all(self, cls: typing.Type[T]) -> typing.List[T]:
        resources: typing.List[T] = []
        class_name = cls.__name__
        for r in self.resources:
            if r["type"] == class_name:
                values = without_keys(r, ["type"])
                resource = cls(**values["data"])
                resources.append(resource)

        return resources


DEFAULT_STORAGE = JSONStorage(resources=[])
